/*
 * File: Snowfall.java
 * ================================================================
 * A program that simulates falling snow on a scene.
 */
import acm.program.*;
import acm.graphics.*;
import acm.util.*;
import java.awt.*;
import java.awt.event.*;

public class Snowfall extends GraphicsProgram {
	/* Preferred window size. */
	public static final int APPLICATION_WIDTH = 500;
	
	/* The delay between frames. */
	private static final double PAUSE_TIME = 1000.0 / 100;
	
	/* How fast the snowflakes fall. */
	private static final double Y_VELOCITY = 5;
	
	/* The color of a snowflake. */
	private static final Color SNOWFLAKE_COLOR = Color.GRAY;
	
	/* The size of a snowflake. */
	private static final double SNOWFLAKE_SIZE = 20;
	
	/* The size of a block. */
	private static final double BLOCK_WIDTH  = 125;
	private static final double BLOCK_HEIGHT = 25;
	
	public void run() {
		addMouseListeners();
		while (true) {
			dropOneSnowflake();
		}
	}
	
	/**
	 * Reacts to a mouse press by drawing a block at that point.
	 */
	public void mousePressed(MouseEvent e) {
		GRect block = new GRect(e.getX() - BLOCK_WIDTH / 2.0,
				                e.getY() - BLOCK_HEIGHT / 2.0,
				                BLOCK_WIDTH, BLOCK_HEIGHT);
		block.setFilled(true);
		add(block);
	}
	
	/**
	 * Given a snowflake, returns whether that snowflake has hit the
	 * bottom of the screen.
	 *
	 * @param snowflake The snowflake in question.
	 * @return Whether it's hit the bottom of the screen.
	 */
	private boolean hasHitBottom(GOval snowflake) {
		double bottomY = snowflake.getY() + snowflake.getHeight();
		return bottomY >= getHeight();
	}
	
	/**
	 * Given a snowflake, generates a random X coordinate at which to
	 * place that snowflake.
	 *
	 * @param snowflake The snowflake in question.
	 * @return A random X coordinate for that snowflake.
	 */
	private double getRandomX(GOval snowflake) {
		RandomGenerator rgen = RandomGenerator.getInstance();
		return rgen.nextDouble(0, getWidth() - snowflake.getWidth());
	}
	
	/**
	 * Creates and returns a new, randomly-positioned snowflake.
	 *
	 * @return A snowflake.
	 */
	private GOval makeSnowflake() {
		GOval result = new GOval(SNOWFLAKE_SIZE, SNOWFLAKE_SIZE);
		result.setFilled(true);
		result.setColor(SNOWFLAKE_COLOR);
		result.setLocation(getRandomX(result), 0);
		return result;
	}
	
	/**
	 * Given a snowflake, returns whether that snowflake has collided
	 * with any other objects on screen.
	 *
	 * @param snowflake The snowflake in question.
	 * @return Whether the snowflake has hit anything else on screen.
	 */
	private boolean hasHitSomethingElse(GOval snowflake) {
		GObject hit =
		  getElementAt(snowflake.getX() + snowflake.getWidth() / 2.0,
				       snowflake.getY() + snowflake.getHeight());
		return hit != snowflake;
	}
	
	/**
	 * Simulates dropping one snowflake on the screen.
	 */
	private void dropOneSnowflake() {
		/* Create and add the snowflake. */
		GOval snowflake = makeSnowflake();
		add(snowflake);
		
		/* To make collision-detection work correctly, send the snowflake 
		 * to the back of the z-order.
		 */
		snowflake.sendToBack();
		
		/* Move the snowflake until it hits something. */
		while (!hasHitBottom(snowflake) && !hasHitSomethingElse(snowflake)) {
			snowflake.move(0, Y_VELOCITY);
			pause(PAUSE_TIME);
		}
	}
}